@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: Hooks.h,v 1.3 1999/02/05 16:02:22 simonm Exp $
+ * $Id: Hooks.h,v 1.5 2001/04/01 05:56:29 chak Exp $
  *
  * (c) The GHC Team, 1998-1999
  *
@@ -17,3 +17,7 @@ extern void PatErrorHdrHook (long fd);
 extern void defaultsHook (void);
 extern void PreTraceHook (long fd);
 extern void PostTraceHook (long fd);
+#if defined(PAR)
+extern void InitEachPEHook (void);
+extern void ShutdownEachPEHook (void);
+#endif
