@@ -1,7 +1,7 @@
 /* 
  * (c) The GRASP/AQUA Project, Glasgow University, 1998
  *
- * $Id: writeError.c,v 1.5 2000/05/01 14:44:25 panne Exp $
+ * $Id: writeError.c,v 1.8 2001/07/16 09:23:24 rrt Exp $
  *
  * hPutStr Runtime Support
  */
@@ -18,29 +18,18 @@ implementation in one or two places.)
 
 #include "Rts.h"
 #include "RtsUtils.h"
-#include "stgio.h"
-
-#ifdef HAVE_FCNTL_H
-#include <fcntl.h>
-#endif
-
-StgAddr
-addrOf_ErrorHdrHook(void)
-{
-  return &ErrorHdrHook;
-}
+#include "HsStd.h"
 
 void
-writeErrString__ (msg_hdr, msg, len)
-StgAddr msg_hdr;
-StgByteArray msg;
-StgInt len;
+writeErrString__ (HsAddr msg_hdr, HsAddr msg, HsInt len)
 {
   int count = 0;
   char* p  = (char*)msg;
   char  nl = '\n';
 
+#ifndef DLLized
   resetNonBlockingFd(2);
+#endif
 
   /* Print error msg header */
   if (msg_hdr) {
