@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: RtsUtils.h,v 1.9.2.1 2000/05/18 14:26:09 simonmar Exp $
+ * $Id: RtsUtils.h,v 1.11.2.1 2001/12/13 14:22:31 simonmar Exp $
  *
  * (c) The GHC Team, 1998-1999
  *
@@ -11,6 +11,8 @@ extern void *stgMallocBytes(int n, char *msg);
 extern void *stgMallocWords(int n, char *msg);
 extern void *stgReallocBytes(void *p, int n, char *msg);
 extern void *stgReallocWords(void *p, int n, char *msg);
+extern void *stgCallocBytes(int n, int m, char *msg);
+
 extern void barf(char *s, ...) __attribute__((__noreturn__)) ;
 extern void belch(char *s, ...);
 extern void prog_belch(char *s, ...);
@@ -19,7 +21,8 @@ extern void _stgAssert (char *filename, unsigned int linenum);
 
 extern void heapOverflow(void);
 
-void resetNonBlockingFd(int fd);
+extern void setNonBlockingFd(int fd);
+extern void resetNonBlockingFd(int fd);
 
 extern nat stg_strlen(char *str);
 
