@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: Stats.h,v 1.8 1999/11/09 15:46:58 simonmar Exp $
+ * $Id: Stats.h,v 1.11 2001/07/23 17:23:20 simonmar Exp $
  *
  * (c) The GHC Team, 1998-1999
  *
@@ -7,12 +7,8 @@
  *
  * ---------------------------------------------------------------------------*/
 
-extern StgDouble elapsedtime(void);
-extern void      start_time(void);
-extern StgDouble usertime(void);
-extern void      end_init(void);
-extern void      stat_exit(int alloc);
-extern void      stat_workerStop(void);
+extern void      stat_startInit(void);
+extern void      stat_endInit(void);
 
 extern void      stat_startGC(void);
 extern void      stat_endGC(lnat alloc, lnat collect, lnat live, 
@@ -21,7 +17,13 @@ extern void      stat_endGC(lnat alloc, lnat collect, lnat live,
 extern void      stat_startExit(void);
 extern void      stat_endExit(void);
 
+extern void      stat_exit(int alloc);
+extern void      stat_workerStop(void);
+
 extern void      initStats(void);
-extern void      stat_describe_gens(void);
+
 extern double    mut_user_time_during_GC(void);
 extern double    mut_user_time(void);
+
+extern void      statDescribeGens( void );
+extern HsInt     getAllocations( void );
