@@ -1,5 +1,5 @@
 /* -----------------------------------------------------------------------------
- * $Id: Storage.c,v 1.25 2000/04/26 11:54:28 simonmar Exp $
+ * $Id: Storage.c,v 1.49.4.3 2001/12/13 14:42:23 simonmar Exp $
  *
  * (c) The GHC Team, 1998-1999
  *
@@ -7,6 +7,7 @@
  *
  * ---------------------------------------------------------------------------*/
 
+#include "PosixSource.h"
 #include "Rts.h"
 #include "RtsUtils.h"
 #include "RtsFlags.h"
@@ -14,7 +15,6 @@
 #include "Hooks.h"
 #include "BlockAlloc.h"
 #include "MBlock.h"
-#include "gmp.h"
 #include "Weak.h"
 #include "Sanity.h"
 
@@ -22,14 +22,11 @@
 #include "Schedule.h"
 #include "StoragePriv.h"
 
-#ifndef SMP
-nat nursery_blocks;		/* number of blocks in the nursery */
-#endif
-
 StgClosure    *caf_list         = NULL;
 
 bdescr *small_alloc_list;	/* allocate()d small objects */
 bdescr *large_alloc_list;	/* allocate()d large objects */
+bdescr *pinned_object_block;    /* allocate pinned objects into this block */
 nat alloc_blocks;		/* number of allocate()d blocks since GC */
 nat alloc_blocks_lim;		/* approximate limit on alloc_blocks */
 
@@ -41,6 +38,8 @@ generation *g0;			/* generation 0, for convenience */
 generation *oldest_gen;		/* oldest generation, for convenience */
 step *g0s0;			/* generation 0, step 0, for convenience */
 
+lnat total_allocated = 0;	/* total memory allocated during run */
+
 /*
  * Storage manager mutex:  protects all the above state from
  * simultaneous access by two STG threads.
@@ -57,16 +56,20 @@ static void *stgReallocForGMP (void *ptr, size_t old_size, size_t new_size);
 static void  stgDeallocForGMP (void *ptr, size_t size);
 
 void
-initStorage (void)
+initStorage( void )
 {
   nat g, s;
-  step *step;
+  step *stp;
   generation *gen;
 
   /* If we're doing heap profiling, we want a two-space heap with a
    * fixed-size allocation area so that we get roughly even-spaced
    * samples.
    */
+
+  /* As an experiment, try a 2 generation collector
+   */
+
 #if defined(PROFILING) || defined(DEBUG)
   if (RtsFlags.ProfFlags.doHeapProfile) {
     RtsFlags.GcFlags.generations = 1;
@@ -76,11 +79,19 @@ initStorage (void)
   }
 #endif
 
-  if (RtsFlags.GcFlags.heapSizeSuggestion > 
+  if (RtsFlags.GcFlags.maxHeapSize != 0 &&
+      RtsFlags.GcFlags.heapSizeSuggestion > 
       RtsFlags.GcFlags.maxHeapSize) {
     RtsFlags.GcFlags.maxHeapSize = RtsFlags.GcFlags.heapSizeSuggestion;
   }
 
+  if (RtsFlags.GcFlags.maxHeapSize != 0 &&
+      RtsFlags.GcFlags.minAllocAreaSize > 
+      RtsFlags.GcFlags.maxHeapSize) {
+      prog_belch("maximum heap size (-M) is smaller than minimum alloc area size (-A)");
+      exit(1);
+  }
+
   initBlockAllocator();
   
   /* allocate generation info array */
@@ -129,19 +140,24 @@ initStorage (void)
   /* Initialise all steps */
   for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
     for (s = 0; s < generations[g].n_steps; s++) {
-      step = &generations[g].steps[s];
-      step->no = s;
-      step->blocks = NULL;
-      step->n_blocks = 0;
-      step->gen = &generations[g];
-      step->hp = NULL;
-      step->hpLim = NULL;
-      step->hp_bd = NULL;
-      step->scan = NULL;
-      step->scan_bd = NULL;
-      step->large_objects = NULL;
-      step->new_large_objects = NULL;
-      step->scavenged_large_objects = NULL;
+      stp = &generations[g].steps[s];
+      stp->no = s;
+      stp->blocks = NULL;
+      stp->n_blocks = 0;
+      stp->gen = &generations[g];
+      stp->gen_no = g;
+      stp->hp = NULL;
+      stp->hpLim = NULL;
+      stp->hp_bd = NULL;
+      stp->scan = NULL;
+      stp->scan_bd = NULL;
+      stp->large_objects = NULL;
+      stp->n_large_blocks = 0;
+      stp->new_large_objects = NULL;
+      stp->scavenged_large_objects = NULL;
+      stp->n_scavenged_large_blocks = 0;
+      stp->is_compacted = 0;
+      stp->bitmap = NULL;
     }
   }
   
@@ -153,8 +169,14 @@ initStorage (void)
     generations[g].steps[s].to = &generations[g+1].steps[0];
   }
   
-  /* The oldest generation has one step and its destination is the
-   * same step. */
+  /* The oldest generation has one step and it is compacted. */
+  if (RtsFlags.GcFlags.compact) {
+      if (RtsFlags.GcFlags.generations == 1) {
+	  belch("WARNING: compaction is incompatible with -G1; disabled");
+      } else {
+	  oldest_gen->steps[0].is_compacted = 1;
+      }
+  }
   oldest_gen->steps[0].to = &oldest_gen->steps[0];
 
   /* generation 0 is special: that's the nursery */
@@ -186,18 +208,54 @@ initStorage (void)
   pthread_mutex_init(&sm_mutex, NULL);
 #endif
 
-  IF_DEBUG(gc, stat_describe_gens());
+  IF_DEBUG(gc, statDescribeGens());
 }
 
 void
 exitStorage (void)
 {
-  stat_exit(calcAllocated());
+    stat_exit(calcAllocated());
 }
 
-
 /* -----------------------------------------------------------------------------
    CAF management.
+
+   The entry code for every CAF does the following:
+     
+      - builds a CAF_BLACKHOLE in the heap
+      - pushes an update frame pointing to the CAF_BLACKHOLE
+      - invokes UPD_CAF(), which:
+          - calls newCaf, below
+	  - updates the CAF with a static indirection to the CAF_BLACKHOLE
+      
+   Why do we build a BLACKHOLE in the heap rather than just updating
+   the thunk directly?  It's so that we only need one kind of update
+   frame - otherwise we'd need a static version of the update frame too.
+
+   newCaf() does the following:
+       
+      - it puts the CAF on the oldest generation's mut-once list.
+        This is so that we can treat the CAF as a root when collecting
+	younger generations.
+
+   For GHCI, we have additional requirements when dealing with CAFs:
+
+      - we must *retain* all dynamically-loaded CAFs ever entered,
+        just in case we need them again.
+      - we must be able to *revert* CAFs that have been evaluated, to
+        their pre-evaluated form.
+
+      To do this, we use an additional CAF list.  When newCaf() is
+      called on a dynamically-loaded CAF, we add it to the CAF list
+      instead of the old-generation mutable list, and save away its
+      old info pointer (in caf->saved_info) for later reversion.
+
+      To revert all the CAFs, we traverse the CAF list and reset the
+      info pointer to caf->saved_info, then throw away the CAF list.
+      (see GC.c:revertCAFs()).
+
+      -- SDM 29/1/01
+
    -------------------------------------------------------------------------- */
 
 void
@@ -212,77 +270,28 @@ newCAF(StgClosure* caf)
    */
   ACQUIRE_LOCK(&sm_mutex);
 
-  ASSERT( ((StgMutClosure*)caf)->mut_link == NULL );
-  ((StgMutClosure *)caf)->mut_link = oldest_gen->mut_once_list;
-  oldest_gen->mut_once_list = (StgMutClosure *)caf;
-
-#ifdef INTERPRETER
-  /* If we're Hugs, we also have to put it in the CAF table, so that
-     the CAF can be reverted.  When reverting, CAFs created by compiled
-     code are recorded in the CAF table, which lives outside the
-     heap, in mallocville.  CAFs created by interpreted code are
-     chained together via the link fields in StgCAFs, and are not
-     recorded in the CAF table.
-  */
-  ASSERT( get_itbl(caf)->type == THUNK_STATIC );
-  addToECafTable ( caf, get_itbl(caf) );
-#endif
+  if (is_dynamically_loaded_rwdata_ptr((StgPtr)caf)) {
+      ((StgIndStatic *)caf)->saved_info  = (StgInfoTable *)caf->header.info;
+      ((StgIndStatic *)caf)->static_link = caf_list;
+      caf_list = caf;
+  } else {
+      ((StgIndStatic *)caf)->saved_info = NULL;
+      ((StgMutClosure *)caf)->mut_link = oldest_gen->mut_once_list;
+      oldest_gen->mut_once_list = (StgMutClosure *)caf;
+  }
 
   RELEASE_LOCK(&sm_mutex);
+
+#ifdef PAR
+  /* If we are PAR or DIST then  we never forget a CAF */
+  { globalAddr *newGA;
+    //belch("<##> Globalising CAF %08x %s",caf,info_type(caf));
+    newGA=makeGlobal(caf,rtsTrue); /*given full weight*/
+    ASSERT(newGA);
+  } 
+#endif /* PAR */
 }
 
-#ifdef INTERPRETER
-void
-newCAF_made_by_Hugs(StgCAF* caf)
-{
-  ACQUIRE_LOCK(&sm_mutex);
-
-  ASSERT( get_itbl(caf)->type == CAF_ENTERED );
-  recordOldToNewPtrs((StgMutClosure*)caf);
-  caf->link = ecafList;
-  ecafList = caf->link;
-
-  RELEASE_LOCK(&sm_mutex);
-}
-#endif
-
-#ifdef INTERPRETER
-/* These initialisations are critical for correct operation
-   on the first call of addToECafTable. 
-*/
-StgCAF*         ecafList      = END_ECAF_LIST;
-StgCAFTabEntry* ecafTable     = NULL;
-StgInt          usedECafTable = 0;
-StgInt          sizeECafTable = 0;
-
-
-void clearECafTable ( void )
-{
-   usedECafTable = 0;
-}
-
-void addToECafTable ( StgClosure* closure, StgInfoTable* origItbl )
-{
-   StgInt          i;
-   StgCAFTabEntry* et2;
-   if (usedECafTable == sizeECafTable) {
-      /* Make the initial table size be 8 */
-      sizeECafTable *= 2;
-      if (sizeECafTable == 0) sizeECafTable = 8;
-      et2 = stgMallocBytes ( 
-               sizeECafTable * sizeof(StgCAFTabEntry),
-               "addToECafTable" );
-      for (i = 0; i < usedECafTable; i++) 
-         et2[i] = ecafTable[i];
-      if (ecafTable) free(ecafTable);
-      ecafTable = et2;
-   }
-   ecafTable[usedECafTable].closure  = closure;
-   ecafTable[usedECafTable].origItbl = origItbl;
-   usedECafTable++;
-}
-#endif
-
 /* -----------------------------------------------------------------------------
    Nursery management.
    -------------------------------------------------------------------------- */
@@ -301,7 +310,7 @@ allocNurseries( void )
       cap->rNursery = allocNursery(NULL, RtsFlags.GcFlags.minAllocAreaSize);
       cap->rCurrentNursery = cap->rNursery;
       for (bd = cap->rNursery; bd != NULL; bd = bd->link) {
-	bd->back = (bdescr *)cap;
+	bd->u.back = (bdescr *)cap;
       }
     }
     /* Set the back links to be equal to the Capability,
@@ -309,12 +318,12 @@ allocNurseries( void )
      */
   }
 #else /* SMP */
-  nursery_blocks  = RtsFlags.GcFlags.minAllocAreaSize;
-  g0s0->blocks    = allocNursery(NULL, nursery_blocks);
-  g0s0->n_blocks  = nursery_blocks;
-  g0s0->to_space  = NULL;
-  MainRegTable.rNursery        = g0s0->blocks;
-  MainRegTable.rCurrentNursery = g0s0->blocks;
+  g0s0->blocks      = allocNursery(NULL, RtsFlags.GcFlags.minAllocAreaSize);
+  g0s0->n_blocks    = RtsFlags.GcFlags.minAllocAreaSize;
+  g0s0->to_blocks   = NULL;
+  g0s0->n_to_blocks = 0;
+  MainCapability.r.rNursery        = g0s0->blocks;
+  MainCapability.r.rCurrentNursery = g0s0->blocks;
   /* hp, hpLim, hp_bd, to_space etc. aren't used in G0S0 */
 #endif
 }
@@ -332,7 +341,7 @@ resetNurseries( void )
   for (cap = free_capabilities; cap != NULL; cap = cap->link) {
     for (bd = cap->rNursery; bd; bd = bd->link) {
       bd->free = bd->start;
-      ASSERT(bd->gen == g0);
+      ASSERT(bd->gen_no == 0);
       ASSERT(bd->step == g0s0);
       IF_DEBUG(sanity,memset(bd->start, 0xaa, BLOCK_SIZE));
     }
@@ -341,45 +350,53 @@ resetNurseries( void )
 #else
   for (bd = g0s0->blocks; bd; bd = bd->link) {
     bd->free = bd->start;
-    ASSERT(bd->gen == g0);
+    ASSERT(bd->gen_no == 0);
     ASSERT(bd->step == g0s0);
     IF_DEBUG(sanity,memset(bd->start, 0xaa, BLOCK_SIZE));
   }
-  MainRegTable.rNursery = g0s0->blocks;
-  MainRegTable.rCurrentNursery = g0s0->blocks;
+  MainCapability.r.rNursery = g0s0->blocks;
+  MainCapability.r.rCurrentNursery = g0s0->blocks;
 #endif
 }
 
 bdescr *
-allocNursery (bdescr *last_bd, nat blocks)
+allocNursery (bdescr *tail, nat blocks)
 {
   bdescr *bd;
   nat i;
 
-  /* Allocate a nursery */
+  // Allocate a nursery: we allocate fresh blocks one at a time and
+  // cons them on to the front of the list, not forgetting to update
+  // the back pointer on the tail of the list to point to the new block.
   for (i=0; i < blocks; i++) {
     bd = allocBlock();
-    bd->link = last_bd;
+    bd->link = tail;
+    // double-link the nursery: we might need to insert blocks
+    if (tail != NULL) {
+	tail->u.back = bd;
+    }
     bd->step = g0s0;
-    bd->gen = g0;
-    bd->evacuated = 0;
+    bd->gen_no = 0;
+    bd->flags = 0;
     bd->free = bd->start;
-    last_bd = bd;
+    tail = bd;
   }
-  return last_bd;
+  tail->u.back = NULL;
+  return tail;
 }
 
 void
 resizeNursery ( nat blocks )
 {
   bdescr *bd;
+  nat nursery_blocks;
 
 #ifdef SMP
   barf("resizeNursery: can't resize in SMP mode");
 #endif
 
+  nursery_blocks = g0s0->n_blocks;
   if (nursery_blocks == blocks) {
-    ASSERT(g0s0->n_blocks == blocks);
     return;
   }
 
@@ -394,15 +411,25 @@ resizeNursery ( nat blocks )
     
     IF_DEBUG(gc, fprintf(stderr, "Decreasing size of nursery to %d blocks\n", 
 			 blocks));
-    for (bd = g0s0->blocks; nursery_blocks > blocks; nursery_blocks--) {
-      next_bd = bd->link;
-      freeGroup(bd);
-      bd = next_bd;
+
+    bd = g0s0->blocks;
+    while (nursery_blocks > blocks) {
+	next_bd = bd->link;
+	next_bd->u.back = NULL;
+	nursery_blocks -= bd->blocks; // might be a large block
+	freeGroup(bd);
+	bd = next_bd;
     }
     g0s0->blocks = bd;
+    // might have gone just under, by freeing a large block, so make
+    // up the difference.
+    if (nursery_blocks < blocks) {
+	g0s0->blocks = allocNursery(g0s0->blocks, blocks-nursery_blocks);
+    }
   }
   
-  g0s0->n_blocks = nursery_blocks = blocks;
+  g0s0->n_blocks = blocks;
+  ASSERT(countBlocks(g0s0->blocks) == g0s0->n_blocks);
 }
 
 /* -----------------------------------------------------------------------------
@@ -414,7 +441,7 @@ resizeNursery ( nat blocks )
    -------------------------------------------------------------------------- */
 
 StgPtr
-allocate(nat n)
+allocate( nat n )
 {
   bdescr *bd;
   StgPtr p;
@@ -430,9 +457,9 @@ allocate(nat n)
     nat req_blocks =  (lnat)BLOCK_ROUND_UP(n*sizeof(W_)) / BLOCK_SIZE;
     bd = allocGroup(req_blocks);
     dbl_link_onto(bd, &g0s0->large_objects);
-    bd->gen  = g0;
+    bd->gen_no  = 0;
     bd->step = g0s0;
-    bd->evacuated = 0;
+    bd->flags = BF_LARGE;
     bd->free = bd->start;
     /* don't add these blocks to alloc_blocks, since we're assuming
      * that large objects are likely to remain live for quite a while
@@ -451,25 +478,85 @@ allocate(nat n)
     bd = allocBlock();
     bd->link = small_alloc_list;
     small_alloc_list = bd;
-    bd->gen = g0;
+    bd->gen_no = 0;
     bd->step = g0s0;
-    bd->evacuated = 0;
+    bd->flags = 0;
     alloc_Hp = bd->start;
     alloc_HpLim = bd->start + BLOCK_SIZE_W;
     alloc_blocks++;
   }
-  
+
   p = alloc_Hp;
   alloc_Hp += n;
   RELEASE_LOCK(&sm_mutex);
   return p;
 }
 
-lnat allocated_bytes(void)
+lnat
+allocated_bytes( void )
 {
   return (alloc_blocks * BLOCK_SIZE_W - (alloc_HpLim - alloc_Hp));
 }
 
+/* ---------------------------------------------------------------------------
+   Allocate a fixed/pinned object.
+
+   We allocate small pinned objects into a single block, allocating a
+   new block when the current one overflows.  The block is chained
+   onto the large_object_list of generation 0 step 0.
+
+   NOTE: The GC can't in general handle pinned objects.  This
+   interface is only safe to use for ByteArrays, which have no
+   pointers and don't require scavenging.  It works because the
+   block's descriptor has the BF_LARGE flag set, so the block is
+   treated as a large object and chained onto various lists, rather
+   than the individual objects being copied.  However, when it comes
+   to scavenge the block, the GC will only scavenge the first object.
+   The reason is that the GC can't linearly scan a block of pinned
+   objects at the moment (doing so would require using the
+   mostly-copying techniques).  But since we're restricting ourselves
+   to pinned ByteArrays, not scavenging is ok.
+
+   This function is called by newPinnedByteArray# which immediately
+   fills the allocated memory with a MutableByteArray#.
+   ------------------------------------------------------------------------- */
+
+StgPtr
+allocatePinned( nat n )
+{
+    StgPtr p;
+    bdescr *bd = pinned_object_block;
+
+    ACQUIRE_LOCK(&sm_mutex);
+    
+    TICK_ALLOC_HEAP_NOCTR(n);
+    CCS_ALLOC(CCCS,n);
+
+    // If the request is for a large object, then allocate()
+    // will give us a pinned object anyway.
+    if (n >= LARGE_OBJECT_THRESHOLD/sizeof(W_)) {
+	RELEASE_LOCK(&sm_mutex);
+	return allocate(n);
+    }
+
+    // If we don't have a block of pinned objects yet, or the current
+    // one isn't large enough to hold the new object, allocate a new one.
+    if (bd == NULL || (bd->free + n) > (bd->start + BLOCK_SIZE_W)) {
+	pinned_object_block = bd = allocBlock();
+	dbl_link_onto(bd, &g0s0->large_objects);
+	bd->gen_no = 0;
+	bd->step   = g0s0;
+	bd->flags  = BF_LARGE;
+	bd->free   = bd->start;
+	alloc_blocks++;
+    }
+
+    p = bd->free;
+    bd->free += n;
+    RELEASE_LOCK(&sm_mutex);
+    return p;
+}
+
 /* -----------------------------------------------------------------------------
    Allocation functions for GMP.
 
@@ -495,7 +582,7 @@ stgAllocForGMP (size_t size_in_bytes)
   
   /* allocate and fill it in. */
   arr = (StgArrWords *)allocate(total_size_in_words);
-  SET_ARR_HDR(arr, &ARR_WORDS_info, CCCS, data_size_in_words);
+  SET_ARR_HDR(arr, &stg_ARR_WORDS_info, CCCS, data_size_in_words);
   
   /* and return a ptr to the goods inside the array */
   return(BYTE_ARR_CTS(arr));
@@ -567,9 +654,9 @@ calcAllocated( void )
   }
 
 #else /* !SMP */
-  bdescr *current_nursery = MainRegTable.rCurrentNursery;
+  bdescr *current_nursery = MainCapability.r.rCurrentNursery;
 
-  allocated = (nursery_blocks * BLOCK_SIZE_W) + allocated_bytes();
+  allocated = (g0s0->n_blocks * BLOCK_SIZE_W) + allocated_bytes();
   for ( bd = current_nursery->link; bd != NULL; bd = bd->link ) {
     allocated -= BLOCK_SIZE_W;
   }
@@ -579,6 +666,7 @@ calcAllocated( void )
   }
 #endif
 
+  total_allocated += allocated;
   return allocated;
 }  
 
@@ -590,10 +678,10 @@ calcLive(void)
 {
   nat g, s;
   lnat live = 0;
-  step *step;
+  step *stp;
 
   if (RtsFlags.GcFlags.generations == 1) {
-    live = (g0s0->to_blocks - 1) * BLOCK_SIZE_W + 
+    live = (g0s0->n_to_blocks - 1) * BLOCK_SIZE_W + 
       ((lnat)g0s0->hp_bd->free - (lnat)g0s0->hp_bd->start) / sizeof(W_);
     return live;
   }
@@ -606,9 +694,12 @@ calcLive(void)
       if (g == 0 && s == 0) { 
 	  continue; 
       }
-      step = &generations[g].steps[s];
-      live += (step->n_blocks - 1) * BLOCK_SIZE_W +
-	((lnat)step->hp_bd->free - (lnat)step->hp_bd->start) / sizeof(W_);
+      stp = &generations[g].steps[s];
+      live += (stp->n_large_blocks + stp->n_blocks - 1) * BLOCK_SIZE_W;
+      if (stp->hp_bd != NULL) {
+	  live += ((lnat)stp->hp_bd->free - (lnat)stp->hp_bd->start) 
+	      / sizeof(W_);
+      }
     }
   }
   return live;
@@ -624,22 +715,25 @@ calcLive(void)
 extern lnat 
 calcNeeded(void)
 {
-  lnat needed = 0;
-  nat g, s;
-  step *step;
-
-  for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
-    for (s = 0; s < generations[g].n_steps; s++) {
-      if (g == 0 && s == 0) { continue; }
-      step = &generations[g].steps[s];
-      if (generations[g].steps[0].n_blocks > generations[g].max_blocks) {
-	needed += 2 * step->n_blocks;
-      } else {
-	needed += step->n_blocks;
-      }
+    lnat needed = 0;
+    nat g, s;
+    step *stp;
+    
+    for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
+	for (s = 0; s < generations[g].n_steps; s++) {
+	    if (g == 0 && s == 0) { continue; }
+	    stp = &generations[g].steps[s];
+	    if (generations[g].steps[0].n_blocks +
+		generations[g].steps[0].n_large_blocks 
+		> generations[g].max_blocks
+		&& stp->is_compacted == 0) {
+		needed += 2 * stp->n_blocks;
+	    } else {
+		needed += stp->n_blocks;
+	    }
+	}
     }
-  }
-  return needed;
+    return needed;
 }
 
 /* -----------------------------------------------------------------------------
@@ -652,11 +746,11 @@ calcNeeded(void)
 
 #ifdef DEBUG
 
-extern void
+void
 memInventory(void)
 {
   nat g, s;
-  step *step;
+  step *stp;
   bdescr *bd;
   lnat total_blocks = 0, free_blocks = 0;
 
@@ -664,13 +758,13 @@ memInventory(void)
 
   for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
     for (s = 0; s < generations[g].n_steps; s++) {
-      step = &generations[g].steps[s];
-      total_blocks += step->n_blocks;
+      stp = &generations[g].steps[s];
+      total_blocks += stp->n_blocks;
       if (RtsFlags.GcFlags.generations == 1) {
 	/* two-space collector has a to-space too :-) */
-	total_blocks += g0s0->to_blocks;
+	total_blocks += g0s0->n_to_blocks;
       }
-      for (bd = step->large_objects; bd; bd = bd->link) {
+      for (bd = stp->large_objects; bd; bd = bd->link) {
 	total_blocks += bd->blocks;
 	/* hack for megablock groups: they have an extra block or two in
 	   the second and subsequent megablocks where the block
@@ -695,45 +789,55 @@ memInventory(void)
   /* count the blocks on the free list */
   free_blocks = countFreeList();
 
-  ASSERT(total_blocks + free_blocks == mblocks_allocated * BLOCKS_PER_MBLOCK);
-
-#if 0
   if (total_blocks + free_blocks != mblocks_allocated *
       BLOCKS_PER_MBLOCK) {
     fprintf(stderr, "Blocks: %ld live + %ld free  = %ld total (%ld around)\n",
 	    total_blocks, free_blocks, total_blocks + free_blocks,
 	    mblocks_allocated * BLOCKS_PER_MBLOCK);
   }
-#endif
+
+  ASSERT(total_blocks + free_blocks == mblocks_allocated * BLOCKS_PER_MBLOCK);
+}
+
+
+nat
+countBlocks(bdescr *bd)
+{
+    nat n;
+    for (n=0; bd != NULL; bd=bd->link) {
+	n += bd->blocks;
+    }
+    return n;
 }
 
 /* Full heap sanity check. */
-
-extern void
-checkSanity(nat N)
+void
+checkSanity( void )
 {
-  nat g, s;
+    nat g, s;
 
-  if (RtsFlags.GcFlags.generations == 1) {
-    checkHeap(g0s0->to_space, NULL);
-    checkChain(g0s0->large_objects);
-  } else {
-    
-    for (g = 0; g <= N; g++) {
-      for (s = 0; s < generations[g].n_steps; s++) {
-	if (g == 0 && s == 0) { continue; }
-	checkHeap(generations[g].steps[s].blocks, NULL);
-      }
+    if (RtsFlags.GcFlags.generations == 1) {
+	checkHeap(g0s0->to_blocks);
+	checkChain(g0s0->large_objects);
+    } else {
+	
+	for (g = 0; g < RtsFlags.GcFlags.generations; g++) {
+	    for (s = 0; s < generations[g].n_steps; s++) {
+		ASSERT(countBlocks(generations[g].steps[s].blocks)
+		       == generations[g].steps[s].n_blocks);
+		ASSERT(countBlocks(generations[g].steps[s].large_objects)
+		       == generations[g].steps[s].n_large_blocks);
+		if (g == 0 && s == 0) { continue; }
+		checkHeap(generations[g].steps[s].blocks);
+		checkChain(generations[g].steps[s].large_objects);
+		if (g > 0) {
+		    checkMutableList(generations[g].mut_list, g);
+		    checkMutOnceList(generations[g].mut_once_list, g);
+		}
+	    }
+	}
+	checkFreeListSanity();
     }
-    for (g = N+1; g < RtsFlags.GcFlags.generations; g++) {
-      for (s = 0; s < generations[g].n_steps; s++) {
-	checkHeap(generations[g].steps[s].blocks,
-		  generations[g].steps[s].blocks->start);
-	checkChain(generations[g].steps[s].large_objects);
-      }
-    }
-    checkFreeListSanity();
-  }
 }
 
 #endif
