@@ -31,7 +31,7 @@ module PatriciaLoMap (
 
 import Prelude hiding (null,map,lookup,foldr,foldl,foldr1,foldl1,filter)
 import EdisonPrelude(Maybe2(..))
-import Assoc ( AssocX(..), Assoc(..), FiniteMapX(..), FiniteMap(..) )
+import qualified Assoc as A ( AssocX(..), Assoc(..), FiniteMapX(..), FiniteMap(..) )
 import qualified Sequence as S
 import AssocDefaults
 import Int
@@ -67,13 +67,13 @@ lowestBit x = x .&. (-x)
 
 branchingBit :: Int -> Int -> Int
 branchingBit p0 p1 = 
-  int32ToInt (lowestBit ((intToInt32 p0) `xor` (intToInt32 p1)))
+  fromIntegral (lowestBit (fromIntegral p0 `xor` fromIntegral p1))
 
 mask :: Int -> Int -> Int
-mask p m = int32ToInt (intToInt32 p .&. (intToInt32 m - 1))
+mask p m = fromIntegral (fromIntegral p .&. (fromIntegral m - (1 :: Int32)))
 
 zeroBit :: Int -> Int -> Bool
-zeroBit p m = (intToInt32 p) .&. (intToInt32 m) == 0
+zeroBit p m = (fromIntegral p) .&. (fromIntegral m) == (0 :: Int32)
 
 matchPrefix :: Int -> Int -> Int -> Bool
 matchPrefix k p m = mask k m == p
@@ -474,33 +474,33 @@ unionSeqWithKey = unionSeqWithKeyUsingReduce
 
 -- instance declarations
 
-instance AssocX FM Int where
-  {empty_ = empty; single_ = single; fromSeq_ = fromSeq; insert_ = insert; 
-   insertSeq_ = insertSeq; union_ = union; unionSeq_ = unionSeq; 
-   delete_ = delete; deleteAll_ = deleteAll; deleteSeq_ = deleteSeq; 
-   null_ = null; size_ = size; member_ = member; count_ = count; 
-   lookup_ = lookup; lookupM_ = lookupM; lookupAll_ = lookupAll; 
-   lookupWithDefault_ = lookupWithDefault; adjust_ = adjust; 
-   adjustAll_ = adjustAll; map_ = map; fold_ = fold; fold1_ = fold1; 
-   filter_ = filter; partition_ = partition; elements_ = elements;
-   instanceName_ m = moduleName}
+instance A.AssocX FM Int where
+  {A.empty = empty; A.single = single; A.fromSeq = fromSeq; A.insert = insert; 
+   A.insertSeq = insertSeq; A.union = union; A.unionSeq = unionSeq; 
+   A.delete = delete; A.deleteAll = deleteAll; A.deleteSeq = deleteSeq; 
+   A.null = null; A.size = size; A.member = member; A.count = count; 
+   A.lookup = lookup; A.lookupM = lookupM; A.lookupAll = lookupAll; 
+   A.lookupWithDefault = lookupWithDefault; A.adjust = adjust; 
+   A.adjustAll = adjustAll; A.map = map; A.fold = fold; A.fold1 = fold1; 
+   A.filter = filter; A.partition = partition; A.elements = elements;
+   A.instanceName m = moduleName}
 
-instance Assoc FM Int where
-  {toSeq_ = toSeq; keys_ = keys; mapWithKey_ = mapWithKey; 
-   foldWithKey_ = foldWithKey; filterWithKey_ = filterWithKey; 
-   partitionWithKey_ = partitionWithKey}
+instance A.Assoc FM Int where
+  {A.toSeq = toSeq; A.keys = keys; A.mapWithKey = mapWithKey; 
+   A.foldWithKey = foldWithKey; A.filterWithKey = filterWithKey; 
+   A.partitionWithKey = partitionWithKey}
 
-instance FiniteMapX FM Int where
-  {fromSeqWith_ = fromSeqWith; fromSeqWithKey_ = fromSeqWithKey; 
-   insertWith_ = insertWith; insertWithKey_ = insertWithKey; 
-   insertSeqWith_ = insertSeqWith; insertSeqWithKey_ = insertSeqWithKey; 
-   unionl_ = unionl; unionr_ = unionr; unionWith_ = unionWith; 
-   unionSeqWith_ = unionSeqWith; intersectWith_ = intersectWith; 
-   difference_ = difference; subset_ = subset; subsetEq_ = subsetEq}
+instance A.FiniteMapX FM Int where
+  {A.fromSeqWith = fromSeqWith; A.fromSeqWithKey = fromSeqWithKey; 
+   A.insertWith = insertWith; A.insertWithKey = insertWithKey; 
+   A.insertSeqWith = insertSeqWith; A.insertSeqWithKey = insertSeqWithKey; 
+   A.unionl = unionl; A.unionr = unionr; A.unionWith = unionWith; 
+   A.unionSeqWith = unionSeqWith; A.intersectWith = intersectWith; 
+   A.difference = difference; A.subset = subset; A.subsetEq = subsetEq}
 
-instance FiniteMap FM Int where
-  {unionWithKey_ = unionWithKey; unionSeqWithKey_ = unionSeqWithKey; 
-   intersectWithKey_ = intersectWithKey}
+instance A.FiniteMap FM Int where
+  {A.unionWithKey = unionWithKey; A.unionSeqWithKey = unionSeqWithKey; 
+   A.intersectWithKey = intersectWithKey}
 
 instance Functor (FM Int) where
   fmap = map
