@@ -1,15 +1,8 @@
 
-#define NON_POSIX_SOURCE
 #include "Rts.h"
-#include "ghcSockets.h"
-#include "stgio.h"
+#include "HsNet.h"
 
-
-#if defined(HAVE_WINSOCK_H) && !defined(__CYGWIN__) && !defined(__CYGWIN32__)
-#define USE_WINSOCK
-#endif
-
-#ifdef USE_WINSOCK
+#if defined(HAVE_WINSOCK_H) && !defined(__CYGWIN__)
 
 static int winsock_inited = 0;
 static int winsock_uninited = 0;
